#include <allegro.h>
#include <fstream>
#include <sstream>
#include <iostream>
#include "map.h"
#include "main.h"
#include "units.h"
#include "globals.h"
#include "controls.h"
#include "special.h"

using namespace std;

int CameraSpeed = 45;
static SAMPLE* Printer = NULL;

// read map from a file, ok=false if something goes wrong
CMap :: CMap(const char* szFileName, bool &ok)
{
    ifstream Input(szFileName);
    Input >> this->iWidth;
    Input >> this->iHeight;
    this->DataMap = new(int*)[iWidth];
    for (int i=0; i<iWidth; i++)
      DataMap[i] = new int[iHeight];
    for (int j=0; j<iHeight; j++)
      for (int i=0; i<iWidth; i++)
        Input >> DataMap[i][j];       // load numbers from map file
    ok = true;
    iGlobalX = 0;
    iGlobalY = 0;  
    ostringstream stmp;    
    for (int i=0; i<MAX_TILE; i++)   // load tile pictures
    {
        stmp.str("");
        stmp << "Tile" << (i+1);
        Tiles[i] = load_bitmap(get_config_string("Tiles",stmp.str().c_str(),"Pics\\tile.pcx"), NULL);        
        if (Tiles[i] == NULL) ok = false;
    }  
    // load mousemap
    MouseMap = load_bitmap(get_config_string("Globals","MouseMap","mousemap.bmp"),NULL);
    if (MouseMap == NULL) ok = false;
    // define special locations
    iHealthCount = get_config_int("Globals","Medikits",0);
    iHealthX = new int[iHealthCount];
    iHealthY = new int[iHealthCount];
    int counter = 0;
    for (int y=0; y<iHeight; y++)
    {
       for (int x=0; x<iWidth; x++)
       {
           switch (DataMap[x][y])
           {
               case SPECIAL_BOOK: iBookX = x;
                                  iBookY = y;
                                  break;
               case SPECIAL_COMPUTER: iComputerX = x;
                                      iComputerY = y;
                                      break;
               case SPECIAL_HEALTH: if (counter < iHealthCount)
                                    {                                        
                                        iHealthX[counter] = x;
                                        iHealthY[counter] = y;
                                        counter++;
                                    }    
                                    break;
           }    
       }        
   }    
   Printer = load_sample(get_config_string("Globals","Printer","printer.wav"));
}
    
CMap :: ~CMap()
{
    destroy_sample(Printer);
    delete iHealthX;
    delete iHealthY;
    for (int i=0; i<iWidth; i++)
      delete DataMap[i];
    delete DataMap;
    destroy_bitmap(MouseMap);
    for (int i=0; i<MAX_TILE; i++)
      destroy_bitmap(Tiles[i]);
}
    
// draw the map to the offscreen
void CMap :: DoDraw()
{
    int SQR_X = SQUARE_WIDTH;
    int SQR_Y = SQUARE_HEIGHT >> 1;   // (divide by 2)
    int px = Player->x + (Player->w >> 1);
    int py = Player->y + Player->h - SQUARE_ELEVATION;
    int ex,ey;  // enemy x, enemy y
    bool efound; // enemy found
    list<CUnit*>::iterator m = Monsters.begin();
    MouseMapping(px,py,px,py);
    int OFFSET_X = 0;    
    int up = 0;
    BITMAP* tile;
    for (int j=0; j<iHeight; j++)
     {
         for (int i=0; i<iWidth; i++)
         {
             // only draw if the player can see it!             
             if ( ((i+1)*SQUARE_WIDTH >= iGlobalX-SQUARE_WIDTH) &&   
                  (i*SQUARE_WIDTH <= iGlobalX + SCREEN_W + SQUARE_WIDTH) &&
                  (j*SQUARE_HEIGHT >> 1 >= iGlobalY-SQUARE_HEIGHT) &&
                  (j*SQUARE_HEIGHT >> 1 <= iGlobalY + SCREEN_H+SQUARE_HEIGHT))
             {                              
                 if ((DataMap[i][j] == 4) || (DataMap[i][j] == 6) ||
                     (DataMap[i][j] == 9))
                     tile = Tiles[0];   // draw grass there                 
                 else
                    tile = Tiles[DataMap[i][j]];                 
                 up = SQUARE_HEIGHT - tile->h;   
                 draw_sprite(Offscreen,tile,i*SQR_X + OFFSET_X - iGlobalX, j*SQR_Y + up - iGlobalY);
                 if ( (px==i) && (py==j) )
                 {
                   Player->DoDraw();   // draw player                   
                   if ( (i == iBookX) && (j == iBookY) )// player wants to read the book
                   {
                       Menu->iState = READING_STATE;
                       BookReader = new CBookReader(15);
                   }
                   else if ( (i == iComputerX) && (j == iComputerY) )  // player uses the computer
                   {                       
                       if ((bBookRead) && (DataMap[i][j] != 0) )
                       {
                           play_sample(Printer,SampleVolume,0,1000,0);
                           // create XO table                                                                 
                           RiddleTable = new CTableGame();
                           Menu->iState = RIDDLE_STATE;
                           DataMap[i][j] = 0;  // no more computer
                           DataMap[i+2][j+1] = 0;
                           DataMap[i+1][j+2] = 0;
                           DataMap[i+1][j+3] = 0;
                       }
                       else
                       {
                           textout(Offscreen,font,"You must read The Book first!",
                                   10,Menu->h + 10,makecol(255,0,0));
                       }                        
                   }
                   else if (DataMap[i][j] == 2)  // player wants to recharge his energy
                   {                       
                       Player->HP += ENERGY_CHARGE;
                       if (Player->HP > PLAYER_MAX_HP) Player->HP = PLAYER_MAX_HP;                                           
                       DataMap[i][j] = 0;  // medikit is used up
                   }  
                 }      
                 else
                 {
                     m = Monsters.begin();
                     efound = false;
                     while ( (!efound) && (m != Monsters.end()) )
                     {
                         ex = (*m)->x + ((*m)->w >> 1);
                         ey = (*m)->y + (*m)->h - SQUARE_ELEVATION;
                         MouseMapping(ex,ey,ex,ey);
                         if ( (ex == i) && (ey == j) )
                         {
                             (*m)->DoDraw();
                             efound = true;
                         }    
                         else m++;
                     }    
                 }    
             }    
         }    
         if (OFFSET_X == 0) 
            OFFSET_X = SQUARE_WIDTH >> 1;
         else
           OFFSET_X = 0;         
     }
     /*
     for (int v=0; v<iHealthCount; v++)
     {
         textprintf(Offscreen,font,iHealthX[v]*SQUARE_WIDTH-iGlobalX,
                    (iHealthY[v]*SQUARE_HEIGHT>>1)-iGlobalY,makecol(255,255,0),"%d,%d",
                    iHealthX[v]*SQUARE_WIDTH,iHealthY[v]*SQUARE_HEIGHT>>1);
     } 
     */   
}

int SteppedOnSpecialLocation(int x, int y)
{
}    
    
void CMap :: DoScroll()
{
    // turned off - player should not be able to scroll by himself    
    /*
    int MAX_WIDTH = iWidth * SQUARE_WIDTH;
    int MAX_HEIGHT = iHeight * (SQUARE_HEIGHT / 2);
    textprintf(Offscreen,font,10,10,0,"X: %d, Y: %d",mouse_x,mouse_y);
    textprintf(Offscreen,font,10,20,0,"Offset X: %d, Offset Y: %d",iGlobalX,iGlobalY);
    textprintf(Offscreen,font,10,30,0,"Map width: %d, Map Height: %d",MAX_WIDTH,MAX_HEIGHT);
    // scrolling by mouse
    if ( (mouse_x < 5) && (iGlobalX > 0) ) iGlobalX -= 30;
    else if ( (mouse_x > SCREEN_W - 5) && (iGlobalX <= MAX_WIDTH - SCREEN_W) ) iGlobalX += 30;
    if ( (mouse_y < 5) && (iGlobalY > 0) ) iGlobalY -= 30;
    else if ( (mouse_y > SCREEN_H - 5) && (iGlobalY <= MAX_HEIGHT - SCREEN_H) ) iGlobalY += 30;
    // scrolling by keyboard
    if ( key[KEY_LEFT] && (iGlobalX > 0) ) iGlobalX -= 30;
    else if ( key[KEY_RIGHT] && (iGlobalX <= MAX_WIDTH - SCREEN_W) ) iGlobalX += 30;
    if ( key[KEY_UP] && (iGlobalY > 0) ) iGlobalY -= 30;
    else if ( key[KEY_DOWN] && (iGlobalY <= MAX_HEIGHT - SCREEN_H) ) iGlobalY += 30;
    */
} 

void CMap :: CameraToCoord(int x, int y)
{
    /*
    if ( ( x < SCREEN_W / 2 ) 
       iGlobalX -= 5;
    else if (x > SCREEN_W / 2)
       iGlobalX += 5;
    if (y < SCREEN_H / 2)
       iGlobalY -= 5;
    else if (y > SCREEN_H / 2)
       iGlobalY += 5;
       */
    if ( (x < SCREEN_W >> 1) && (x+CameraSpeed < SCREEN_W >> 1) )
       iGlobalX -= CameraSpeed;
    else if ( (x > SCREEN_W >> 1) && (x+CameraSpeed > SCREEN_W >> 1) )
       iGlobalX += CameraSpeed;
    if ( (y < SCREEN_H >> 1) && (y+CameraSpeed < SCREEN_H >> 1) )
       iGlobalY -= CameraSpeed;
    else if ( (y > SCREEN_H >> 1) && (y+CameraSpeed > SCREEN_H >> 1) )
       iGlobalY += CameraSpeed;
}       

// can a unit pass through this terrain or not
inline bool Unpassable(int iTerrainType)
{
    switch (iTerrainType)
    {
        case 0: return false;
        case 5: return true;
        case 7: return true;
        default: return false;
    }    
}   

inline void CMap :: MouseMapping(int X, int Y, int &iTileX, int &iTileY)
{
    int iRegionX = X / MOUSEMAP_WIDTH;
    int iRegionY = (Y / MOUSEMAP_HEIGHT) << 1;
    int iMouseMapX = X % MOUSEMAP_WIDTH;
    int iMouseMapY = Y % MOUSEMAP_HEIGHT;
    /*  yellow = R: 255, G: 255, B: 0
        red = R: 255, G: 0, B: 0
        green = R: 0, G: 255, B: 0
        blue = R: 0, G: 0, B: 0  
        white = R: 255, G: 255, B: 255   */
    int iPixel = getpixel(MouseMap,iMouseMapX,iMouseMapY);
    int R = getr(iPixel);
    int G = getg(iPixel);
    int B = getb(iPixel);
    int iRegionDX, iRegionDY;
    if ( (R == 255) && (G == 255) && (B == 255) )    // white
    {
        iRegionDX = 0;
        iRegionDY = 0;
    }    
    else if ( (R == 255) && (G == 0) && (B == 0) )   // red
    {
       iRegionDX = -1;
       iRegionDY = -1;    
    }    
    else if ( (R == 255) && (G == 255) && (B == 0) ) // yellow
    {
        iRegionDX = 0;
        iRegionDY = -1;
    }    
    else if ( (R == 0) && (G == 255) && (B == 0) )   // green
    {    
        iRegionDX = -1;
        iRegionDY = 1;
    }
    else if ( (R == 0) && (G == 0) && (B == 255) )   // blue
    {
        iRegionDX = 0;
        iRegionDY = 1;
    }
    else       // error
    {
        iRegionDX = 0;
        iRegionDY = 0;        
    }   
    iTileX = iRegionX + iRegionDX;
    iTileY = iRegionY + iRegionDY;
}     

bool CMap :: Collision(int &x, int &y, int w, int h)
{    
    int X = x + 5;   // +5 = make the collision a bit "smoother"
    int Y = y + h - (SQUARE_HEIGHT >> 1);
    int iTileX,iTileY;
    MouseMapping(X,Y,iTileX,iTileY);
    //return Unpassable(DataMap[iTileX][iTileY]);
    
    if (!Unpassable(DataMap[iTileX][iTileY]))
    {
        X = x+w - 5;
        MouseMapping(X,Y,iTileX,iTileY);
        return Unpassable(DataMap[iTileX][iTileY]);
    }
    else
    {
        return true;
    } 
       
}    
